#include <ESP8266httpUpdate.h>
#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include <SoftwareSerial.h>
#include <DFRobotDFPlayerMini.h>
#include <WiFiUdp.h>
#include <NTPClient.h>

#include <YahooWeather.h>
#include <Weather.h>
#include <GmailNotifier.h>
#include <PAConfig.h>
#include <PersonalAssistant.h>

#define BUTTON 5
bool buttonPressed = false;

SoftwareSerial softSerial(4, 2); // Rx, Tx
ESP8266WebServer server(80);
WiFiUDP ntpUDP;
NTPClient timeClient(ntpUDP, (char*)"pool.ntp.org");
YahooWeather yahoo;
Weather weather;
PAConfig config;
DFRobotDFPlayerMini dfPlayer;
PersonalAssistant pa(dfPlayer);
GmailNotifier gmail;
uint8_t buffer[12];


void setup() {
  pinMode(BUTTON, INPUT);
  SPIFFS.begin();
  Serial.begin(115200);
  softSerial.begin(9600);
  Serial.println();
  Serial.println(F("Personal Assistant Started."));

  if (!dfPlayer.begin(softSerial)) {
    Serial.println(F("Unable to begin:"));
    Serial.println(F("1.Please recheck the connection!"));
    Serial.println(F("2.Please insert the SD card!"));
    while(true);
  }
  Serial.println(F("(( DFPlayer Mini Online ))"));

  unsigned long seed = 0;
  for (int i = 0; i < 32; i++) {
    seed = seed | ((analogRead(A0) & 0x01) << i);
  }
  randomSeed(seed);

  dfPlayer.setTimeOut(500); // dfPlayer Serial Communication Timeout
//  dfPlayer.volume(24); // 0 ~ 30
  dfPlayer.volume(28); // 0 ~ 30
//  dfPlayer.volume(16); // 0 ~ 30
  dfPlayer.playFolder(1, random(1, 6));
  delay(3500);

  config.load(SPIFFS.open("/config.json", "r"));
  pa.loadServices(SPIFFS.open(F("/services.json"), "r"));

  yahoo.setUnit(YahooWeather::UNIT_METRIC);
  yahoo.setFormat(YahooWeather::FORMAT_JSON);
  yahoo.setWoeid(config.weather.woeid.c_str());
  yahoo.setAppId(config.weather.appId.c_str());
  yahoo.setConsumerKey(config.weather.consumerKey.c_str());
  yahoo.setConsumerSecret(config.weather.consumerSecret.c_str());
  yahoo.setUrl(config.weather.api.c_str());

  uint8_t i = 0;
  WiFi.hostname("PersonalAssistant");
//  WiFi.setAutoConnect(false);
  WiFi.softAPdisconnect();
  WiFi.disconnect();
  WiFi.mode(WIFI_STA);
  if (WiFi.status() != WL_CONNECTED) {
    WiFi.begin(config.wifi.ssid.c_str(), config.wifi.password.c_str());
    while (WiFi.status() != WL_CONNECTED) {
      Serial.print('.');
      delay(250);
      if (i++ > 80) break;
    }
    if(WiFi.status() == WL_CONNECTED) {
        Serial.print("\nConnected to ");
        Serial.println(WiFi.SSID());
        Serial.print("IP address:\t");
        Serial.println(WiFi.localIP());
    }
    Serial.println();
  } else {
    Serial.println("#wifi 4");
    Serial.print("\nConnected to ");
    Serial.println(WiFi.SSID());
    Serial.print("IP address:\t");
    Serial.println(WiFi.localIP());
  }

  server.on(F("/update"), update);
  server.on(F("/spiffs"), spiffs);
  server.on(F("/api"), api);
  server.onNotFound([]() {                              // If the client requests any URI
    if(!handleFileRead(server.uri()))                   // send it if it exists
      handleFileRead(F("/404.html"));
  });
  server.begin();

  yahoo.request();
  weather.parse(yahoo.response.c_str());

  gmail.setUsername(config.gmail.username.c_str());
  gmail.setPassword(config.gmail.password.c_str());
  gmail.sync();
  Serial.print("gmail unread: ");
  Serial.println(gmail.getUnread());

  timeClient.setTimeOffset(config.time.timezone * 4600);
  timeClient.begin();
}

void loop() {
  timeClient.update();
  server.handleClient();
  pa.handleSpeak();
  if (digitalRead(BUTTON) == HIGH) {
    delay(100);
    if (!buttonPressed) {
      buttonPressed = true;
      switch (pa.getQueueTurn()) {
        case GMAIL_UNREADMESSAGES : { // unreadMessages
          pa.speakQClear();
          if (gmail.getUnread() > 0) {
            pa.numberToWord(gmail.getUnread(), buffer);
            pa.speakQAdd(200, 750);
            for (uint8_t i=1; i<buffer[0]; i++) {
              pa.speakQAdd(buffer[i], PersonalAssistant::calcNumSpeakDuration(buffer[i]) * 10);
            }
            pa.speakQAdd(201, 2500);
          } else if (gmail.getUnread() < 0) {
            pa.speakQAdd(203, 2120); // syncing gmail inbox failed
          } else {
            pa.speakQAdd(202, 1520); // you have no new messages
          }
        }; break;
        case BIRTHDEATH_WORLDBIRTH : { // worldBirth
          pa.speakQClear();
          pa.numberToWord(calcBirth((timeClient.getHours() * 3600) + (timeClient.getMinutes() * 60)), buffer);
          for (uint8_t i=1; i<buffer[0]; i++) {
            pa.speakQAdd(buffer[i], PersonalAssistant::calcNumSpeakDuration(buffer[i]) * 10);
          }
          pa.speakQAdd(219, 2800); // people was born since the beginning of the day
          pa.speakQAdd(221, 2100); // and each blink indicates 10 new
        }; break;
        case BIRTHDEATH_WORLDDEATH : { // worldDeath
          pa.speakQClear();
          pa.numberToWord(calcDeath((timeClient.getHours() * 3600) + (timeClient.getMinutes() * 60)), buffer);
          for (uint8_t i=1; i<buffer[0]; i++) {
            pa.speakQAdd(buffer[i], PersonalAssistant::calcNumSpeakDuration(buffer[i]) * 10);
          }
          pa.speakQAdd(220, 2800); // people was died since the beginning of the day
          pa.speakQAdd(221, 2100); // and each blink indicates 10 new
        }; break;
        case WEATHER_CURRENTTEMPERATURE : { // currentTemperature
          pa.speakQClear();
          pa.speakQAdd(204, 500); // it's
          pa.numberToWord(weather.getConditionTemp(), buffer);
          for (uint8_t i=1; i<buffer[0]; i++) {
            pa.speakQAdd(buffer[i], PersonalAssistant::calcNumSpeakDuration(buffer[i]) * 10);
          }
          pa.speakQAdd(205, 1310); // degrees Celsius
        }; break;
        case WEATHER_TODAYCONDITION : { // todayCondition
          pa.speakQClear();
          Serial.println("@queue today condition");
        }; break;
        case WEATHER_TODAYMINMAX : { // todayMinMax
          pa.speakQClear();
          pa.speakQAdd(206, 560); // today
          pa.speakQAdd(208, 925); // the low will be
          pa.numberToWord(weather.getTodayLow(), buffer);
          for (uint8_t i=1; i<buffer[0]; i++) {
            pa.speakQAdd(buffer[i], PersonalAssistant::calcNumSpeakDuration(buffer[i]) * 10); // today low
          }
          pa.speakQAdd(205, 1310 + 200); // degrees Celsius
          pa.speakQAdd(192, 500); // and
          pa.speakQAdd(209, 950); // the high will be
          pa.numberToWord(weather.getTodayHigh(), buffer);
          for (uint8_t i=1; i<buffer[0]; i++) {
            pa.speakQAdd(buffer[i], PersonalAssistant::calcNumSpeakDuration(buffer[i]) * 10); // today high
          }
          pa.speakQAdd(205, 1310); // degrees Celsius
        }; break;
        case WEATHER_TOMORROWCONDITION : { // tomorrowCondition
          pa.speakQClear();
          switch (Weather::getWeatherCondition(weather.getTodayLow(), weather.getTodayHigh(), weather.getTomorrowLow(), weather.getTomorrowHigh())) {
            case Weather::MODE_WARMER : pa.speakQAdd(225, 1900);
              break;
            case Weather::MODE_COLDER : pa.speakQAdd(226, 1900);
              break;
            case Weather::MODE_NEUTRAL : pa.speakQAdd(227, 2700);
              break;
          }
        }; break;
        case WEATHER_TOMORROWMINMAX : { // tomorrowMinMax
          pa.speakQClear();
          pa.speakQAdd(207, 680); // tomorrow
          pa.speakQAdd(208, 925); // the low will be
          pa.numberToWord(weather.getTomorrowLow(), buffer);
          for (uint8_t i=1; i<buffer[0]; i++) {
            pa.speakQAdd(buffer[i], PersonalAssistant::calcNumSpeakDuration(buffer[i]) * 10); // tomorrow low
          }
          pa.speakQAdd(205, 1310 + 200); // degrees Celsius
          pa.speakQAdd(192, 500); // and
          pa.speakQAdd(209, 950); // the high will be
          pa.numberToWord(weather.getTomorrowHigh(), buffer);
          for (uint8_t i=1; i<buffer[0]; i++) {
            pa.speakQAdd(buffer[i], PersonalAssistant::calcNumSpeakDuration(buffer[i]) * 10); // tomorrow high
          }
          pa.speakQAdd(205, 1310); // degrees Celsius
        }; break;
        case WEATHER_PRECIPITATIONFORECAST : { // precipitationForecast
          pa.speakQClear();
            if (isRain(weather.getTomorrowCode())) {
              switch (random(1, 5)) {
                case 1 : pa.speakQAdd(210, 2940); // tut, tut, looks like rain for tomorrow
                  break;
                case 2 : pa.speakQAdd(211, 2800); // bring your own sunshine, it's rainy tomorrow
                  break;
                case 3 : pa.speakQAdd(213, 2230); // there is possibility for rain tomorrow
                  break;
                case 4 : {
                  uint8_t dice = random(1, 100);
                  pa.speakQAdd(214, 2010); // I guess the sky will cry tomorrow
                  if (dice < 30) { // 30 percent chance to say
                    pa.speakQAdd(215, 1000); // take an umbrella
                  }
                }
                  break;
                case 5 : {
                  uint8_t dice = random(1, 100);
                  pa.speakQAdd(218, 1000); // it's gonna rain tomorrow
                  if (dice < 45) {
                    pa.speakQAdd(217, 1500); // don't forget your umbrella
                  } else if (dice < 70) {
                    pa.speakQAdd(215, 1000); // take an umbrella
                  }
                }
                  break;
              }
            } else if(isSnow(weather.getTomorrowCode())) { // it's snow
              switch (random(1, 2)) {
                case 1 : pa.speakQAdd(212, 2310); // there is possibility for snow tomorrow
                  break;
                case 2 : pa.speakQAdd(216, 1920); // you can make a snowman tomorrow
                  break;
              }
            } else { // no precipitation
              pa.speakQAdd(224, 2600); // there is no chance of precipitation tomorrow
            }
        }; break;
        case WEATHER_SUNRISESUNSET : { // sunriseSunset
          pa.speakQClear();
          pa.speakQAdd(222, 1180); // sunrise is at
          pa.numberToWord(Weather::getHours(weather.getSunrise()), buffer);
          for (uint8_t i=1; i<buffer[0]; i++) {
            pa.speakQAdd(buffer[i], PersonalAssistant::calcNumSpeakDuration(buffer[i]) * 10);
          }
          pa.numberToWord(Weather::getMinutes(weather.getSunrise()), buffer);
          for (uint8_t i=1; i<buffer[0]; i++) {
            pa.speakQAdd(buffer[i], PersonalAssistant::calcNumSpeakDuration(buffer[i]) * 10);
          }
          pa.speakQAdd(192, 500); // and
          pa.speakQAdd(223, 1250); // sunset is at
          pa.numberToWord(Weather::getHours(weather.getSunset()), buffer);
          for (uint8_t i=1; i<buffer[0]; i++) {
            pa.speakQAdd(buffer[i], PersonalAssistant::calcNumSpeakDuration(buffer[i]) * 10);
          }
          pa.numberToWord(Weather::getMinutes(weather.getSunset()), buffer);
          for (uint8_t i=1; i<buffer[0]; i++) {
            pa.speakQAdd(buffer[i], PersonalAssistant::calcNumSpeakDuration(buffer[i]) * 10);
          }
        }; break;
        case TIME_CURRENTTIME : { // currentTime
          pa.speakQClear();
          pa.speakQAdd(204, 500); // it's
          pa.numberToWord(timeClient.getHours(), buffer);
          for (uint8_t i=1; i<buffer[0]; i++) {
            pa.speakQAdd(buffer[i], PersonalAssistant::calcNumSpeakDuration(buffer[i]) * 10);
          }
          pa.numberToWord(timeClient.getMinutes(), buffer);
          for (uint8_t i=1; i<buffer[0]; i++) {
            pa.speakQAdd(buffer[i], PersonalAssistant::calcNumSpeakDuration(buffer[i]) * 10);
          }
        }; break;
        case TIME_CURRENTDATE : { // currentDate
          pa.speakQClear();
          uint16_t date[3];
          uint8_t monthNameDuration[12] = { 95, 80, 57, 57, 52, 48, 52, 66, 73, 73, 71, 70 };
          PersonalAssistant::getDate(timeClient.getEpochTime(), date);
          pa.speakQAdd(204, 500); // it's
          pa.speakQAdd(250 + date[1], monthNameDuration[date[1]] * 10); // [month name]
          pa.numberToWord(date[0], buffer);
          for (uint8_t i=1; i<buffer[0]; i++) {
            pa.speakQAdd(buffer[i], PersonalAssistant::calcNumSpeakDuration(buffer[i]) * 10);
          }
          pa.numberToWord(date[2], buffer);
          for (uint8_t i=1; i<buffer[0]; i++) {
            pa.speakQAdd(buffer[i], PersonalAssistant::calcNumSpeakDuration(buffer[i]) * 10);
          }
        }; break;
      }
    }
  } else {
    if (buttonPressed)
      buttonPressed = false;
  }
}

String getContentType(String filename) { // convert the file extension to the MIME type
  if (filename.endsWith(F(".html"))) return F("text/html; charset=utf-8");
  else if (filename.endsWith(F(".css"))) return F("text/css");
  else if (filename.endsWith(F(".js"))) return F("application/javascript");
  else if (filename.endsWith(F(".ico"))) return F("image/x-icon");
  else if (filename.endsWith(F(".svg"))) return F("image/svg+xml");
  else if (filename.endsWith(F(".json"))) return F("application/json");
  else if (filename.endsWith(F(".xml"))) return F("text/xml");
  return ("text/plain");
}

bool handleFileRead(String path) { // send the right file to the client (if it exists)
  if (path.endsWith(F("/"))) path += F("index.html");      // If a folder is requested, send the index file
  String contentType = getContentType(path);               // Get the MIME type
  if (SPIFFS.exists(path) && path != F("/config.json")) {  // If the file exists
    if (path.endsWith(F(".css")) || path.endsWith(F(".js")) || path.endsWith(F(".ico")) || path.endsWith(F(".svg")) || path.endsWith(F(".eot")) || path.endsWith(F(".woff")) || path.endsWith(F(".ttf")) || path.endsWith(F(".woff2")) || path.endsWith(F(".xml"))) {
      server.sendHeader(F("Cache-Control"), F("private, max-age=31556926"), true);
    }
    File file = SPIFFS.open(path, "r");                 // Open it
    size_t sent = server.streamFile(file, contentType); // And send it to the client
    file.close();                                       // Then close the file again
    return true;
  }
  server.send(200, F("text/plain"), F("HTTP 404 - File Not Found"));
  return false;                                         // If the file doesn't exist, return false
}

void api() {
  uint8_t responseCode = 403;
  if (server.hasArg(F("end"))) {
    String response((const char *)0);
    response.reserve(32);
    response += "{\"end\":\"";
    if (server.arg(F("end")) == F("module")) {
      if (server.hasArg(F("module")) && server.hasArg(F("action"))) {
        bool enable = server.arg(F("action")) == F("disable") ? false : true;
        if (server.arg("module") == F("gmail.unreadMessages")) {
          pa.setModule(GMAIL_UNREADMESSAGES, enable);
        } else if (server.arg("module") == F("birthDeath.worldBirth")) {
          pa.setModule(BIRTHDEATH_WORLDBIRTH, enable);
        } else if (server.arg("module") == F("birthDeath.worldDeath")) {
          pa.setModule(BIRTHDEATH_WORLDDEATH, enable);
        } else if (server.arg("module") == F("weather.currentTemperature")) {
          pa.setModule(WEATHER_CURRENTTEMPERATURE, enable);
        } else if (server.arg("module") == F("weather.todayCondition")) {
          pa.setModule(WEATHER_TODAYCONDITION, enable);
        } else if (server.arg("module") == F("weather.todayMinMax")) {
          pa.setModule(WEATHER_TODAYMINMAX, enable);
        } else if (server.arg("module") == F("weather.tomorrowCondition")) {
          pa.setModule(WEATHER_TOMORROWCONDITION, enable);
        } else if (server.arg("module") == F("weather.tomorrowMinMax")) {
          pa.setModule(WEATHER_TOMORROWMINMAX, enable);
        } else if (server.arg("module") == F("weather.precipitationForecast")) {
          pa.setModule(WEATHER_PRECIPITATIONFORECAST, enable);
        } else if (server.arg("module") == F("weather.sunriseSubset")) {
          pa.setModule(WEATHER_SUNRISESUNSET, enable);
        } else if (server.arg("module") == F("time.currentTime")) {
          pa.setModule(TIME_CURRENTTIME, enable);
        } else if (server.arg("module") == F("time.currentDate")) {
          pa.setModule(TIME_CURRENTDATE, enable);
        }
        pa.saveServices(SPIFFS.open("/services.json", "w"));
        responseCode = 200;
        response += "module";
      }
    } else if (server.arg(F("end")) == F("config")) {
      if (server.hasArg(F("wifi.ssid"))) {
        config.wifi.ssid = server.arg(F("wifi.ssid"));
      }
      if (server.hasArg(F("wifi.password"))) {
        config.wifi.password = server.arg(F("wifi.password"));
      }
      if (server.hasArg(F("gmail.username"))) {
        config.gmail.username = server.arg(F("gmail.username"));
      }
      if (server.hasArg(F("gmail.password"))) {
        config.gmail.password = server.arg(F("gmail.password"));
      }
      if (server.hasArg(F("weather.woeid"))) {
        config.weather.woeid = server.arg(F("weather.woeid")).toInt();
      }
      if (server.hasArg(F("weather.api"))) {
        config.weather.api = server.arg(F("weather.api"));
      }
      if (server.hasArg(F("weather.consumerKey"))) {
        config.weather.consumerKey = server.arg(F("weather.consumerKey"));
      }
      if (server.hasArg(F("weather.consumerSecret"))) {
        config.weather.consumerSecret = server.arg(F("weather.consumerSecret"));
      }
      if (server.hasArg(F("time.server"))) {
        config.time.server = server.arg(F("time.server"));
      }
      if (server.hasArg(F("time.timezone"))) {
        config.time.timezone = server.arg(F("time.timezone")).toFloat();
      }
      config.save(SPIFFS.open("/config.json", "w"));
      responseCode = 200;
      response += "config";
    } else {
      responseCode = 404;
      response += "undef";
    }
    response += "\",\"status\":true}";
    server.send(200, F("application/json"), response);
    return;
  }
  server.send(responseCode);
}

void update() {
  ESPhttpUpdate.update("http://192.168.1.101/personal-assistant/update/sketch.bin");
  server.send(200, F("text/plain"), F("update page"));
}

void spiffs() {
  ESPhttpUpdate.updateSpiffs("http://192.168.1.101/personal-assistant/update/spiffs.bin");
  server.send(200, F("text/plain"), F("spiffs update page"));
}

